# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging

logger = logging.getLogger('pyomo.core')

from pyomo.common.collections import ComponentMap, ComponentSet
from pyomo.common.log import is_debug_set
import pyomo.core.expr as EXPR
from pyomo.core.base import (
    Transformation,
    TransformationFactory,
    Connector,
    Constraint,
    ConstraintList,
    Var,
    SortComponents,
)
from pyomo.core.base.connector import ConnectorData, ScalarConnector


@TransformationFactory.register(
    'core.expand_connectors',
    doc="Expand all connectors in the model to simple constraints",
)
class ExpandConnectors(Transformation):
    def _apply_to(self, instance, **kwds):
        if is_debug_set(logger):  # pragma:nocover
            logger.debug("Calling ConnectorExpander")

        connectorsFound = False
        for c in instance.component_data_objects(Connector):
            connectorsFound = True
            break
        if not connectorsFound:
            return

        if is_debug_set(logger):  # pragma:nocover
            logger.debug("   Connectors found!")

        #
        # At this point, there are connectors in the model, so we must
        # look for constraints that involve connectors and expand them.
        #
        # List of the connectors in the order in which we found them
        # (this should be deterministic, provided that the user's model
        # is deterministic)
        connector_list = []
        # list of constraints with connectors: tuple(constraint, connector_set)
        # (this should be deterministic, provided that the user's model
        # is deterministic)
        constraint_list = []
        # ID of the next connector group (set of matched connectors)
        groupID = 0
        # connector_groups stars out as a dict of {id(set): (groupID, set)}
        # If you sort by the groupID, then this will be deterministic.
        connector_groups = dict()
        # map of connector to the set of connectors that must match it
        matched_connectors = ComponentMap()
        # The set of connectors found in the current constraint
        found = ComponentSet()

        connector_types = set([ScalarConnector, ConnectorData])
        for constraint in instance.component_data_objects(
            Constraint, sort=SortComponents.deterministic
        ):
            ref = None
            for c in EXPR.identify_components(constraint.body, connector_types):
                found.add(c)
                if c in matched_connectors:
                    if ref is None:
                        # The first connector in this constraint has
                        # already been seen.  We will use that Set as
                        # the reference
                        ref = matched_connectors[c]
                    elif ref is not matched_connectors[c]:
                        # We already have a reference group; merge this
                        # new group into it.
                        #
                        # Optimization: this merge is linear in the size
                        # of the src set.  If the reference set is
                        # smaller, save time by switching to a new
                        # reference set.
                        src = matched_connectors[c]
                        if len(ref) < len(src):
                            ref, src = src, ref
                        ref.update(src)
                        for _ in src:
                            matched_connectors[_] = ref
                        del connector_groups[id(src)]
                    # else: pass
                    #   The new group *is* the reference group;
                    #   there is nothing to do.
                else:
                    # The connector has not been seen before.
                    connector_list.append(c)
                    if ref is None:
                        # This is the first connector in the constraint:
                        # start a new reference set.
                        ref = ComponentSet()
                        connector_groups[id(ref)] = (groupID, ref)
                        groupID += 1
                    # This connector hasn't been seen.  Record it.
                    ref.add(c)
                    matched_connectors[c] = ref
            if ref is not None:
                constraint_list.append((constraint, found))
                found = ComponentSet()

        # Validate all connector sets and expand the empty ones
        known_conn_sets = {}
        for groupID, conn_set in sorted(connector_groups.values()):
            known_conn_sets[id(conn_set)] = self._validate_and_expand_connector_set(
                conn_set
            )

        # Expand each constraint
        for constraint, conn_set in constraint_list:
            cList = ConstraintList()
            constraint.parent_block().add_component(
                '%s.expanded' % (constraint.getname(fully_qualified=False),), cList
            )
            connId = next(iter(conn_set))
            ref = known_conn_sets[id(matched_connectors[connId])]
            for k, v in sorted(ref.items()):
                if v[1] >= 0:
                    _iter = v[0]
                else:
                    _iter = (v[0],)
                for idx in _iter:
                    substitution = {}
                    for c in conn_set:
                        if v[1] >= 0:
                            new_v = c.vars[k][idx]
                        elif k in c.aggregators:
                            new_v = c.vars[k].add()
                        else:
                            new_v = c.vars[k]
                        substitution[id(c)] = new_v
                    cList.add(
                        (
                            constraint.lower,
                            EXPR.clone_expression(constraint.body, substitution),
                            constraint.upper,
                        )
                    )
            constraint.deactivate()

        # Now, go back and implement VarList aggregators
        for conn in connector_list:
            block = conn.parent_block()
            for var, aggregator in conn.aggregators.items():
                c = Constraint(expr=aggregator(block, conn.vars[var]))
                block.add_component(
                    '%s.%s.aggregate' % (conn.getname(fully_qualified=True), var), c
                )

    def _validate_and_expand_connector_set(self, connectors):
        ref = {}
        # First, go through the connectors and get the superset of all fields
        for c in connectors:
            for k, v in c.vars.items():
                if k in ref:
                    # We have already seen this var
                    continue
                if v is None:
                    # This is an implicit var
                    continue
                # OK: New var, so add it to the reference list
                _len = (
                    # -3 if v is None else
                    -2
                    if k in c.aggregators
                    else (
                        -1
                        if not hasattr(v, 'is_indexed') or not v.is_indexed()
                        else len(v)
                    )
                )
                ref[k] = (v, _len, c)

        if not ref:
            logger.warning(
                "Cannot identify a reference connector: no connectors "
                "in the connector set have assigned variables:\n\t(%s)"
                % (', '.join(sorted(c.name for c in connectors)),)
            )
            return ref

        # Now make sure that connectors match
        empty_or_partial = []
        for c in connectors:
            c_is_partial = False
            if not c.vars:
                # This is an empty connector and should be defined with
                # "auto" vars
                empty_or_partial.append(c)
                continue

            for k, v in ref.items():
                if k not in c.vars:
                    raise ValueError(
                        "Connector mismatch: Connector '%s' missing variable "
                        "'%s' (appearing in reference connector '%s')"
                        % (c.name, k, v[2].name)
                    )
                _v = c.vars[k]
                if _v is None:
                    if not c_is_partial:
                        empty_or_partial.append(c)
                        c_is_partial = True
                    continue
                _len = (
                    -3
                    if _v is None
                    else (
                        -2
                        if k in c.aggregators
                        else (
                            -1
                            if not hasattr(_v, 'is_indexed') or not _v.is_indexed()
                            else len(_v)
                        )
                    )
                )
                if (_len >= 0) ^ (v[1] >= 0):
                    raise ValueError(
                        "Connector mismatch: Connector variable '%s' mixing "
                        "indexed and non-indexed targets on connectors '%s' "
                        "and '%s'" % (k, v[2].name, c.name)
                    )
                if _len >= 0 and _len != v[1]:
                    raise ValueError(
                        "Connector mismatch: Connector variable '%s' index "
                        "mismatch (%s elements in reference connector '%s', "
                        "but %s elements in connector '%s')"
                        % (k, v[1], v[2].name, _len, c.name)
                    )
                if v[1] >= 0 and len(v[0].index_set() ^ _v.index_set()):
                    raise ValueError(
                        "Connector mismatch: Connector variable '%s' has "
                        "mismatched indices on connectors '%s' and '%s'"
                        % (k, v[2].name, c.name)
                    )

        # as we are adding things to the model, sort by key so that
        # the order things are added is deterministic
        sorted_refs = sorted(ref.items())
        if len(empty_or_partial) > 1:
            # This is expensive (names aren't cheap), but does result in
            # a deterministic ordering
            empty_or_partial.sort(key=lambda x: x.getname(fully_qualified=True))

        # Fill in any empty connectors
        for c in empty_or_partial:
            block = c.parent_block()
            for k, v in sorted_refs:
                if k in c.vars and c.vars[k] is not None:
                    continue

                if v[1] >= 0:
                    idx = (v[0].index_set(),)
                else:
                    idx = ()
                var_args = {}
                try:
                    var_args['domain'] = v[0].domain
                except AttributeError:
                    pass
                try:
                    var_args['bounds'] = v[0].bounds
                except AttributeError:
                    pass
                new_var = Var(*idx, **var_args)
                block.add_component(
                    '%s.auto.%s' % (c.getname(fully_qualified=True), k), new_var
                )
                if idx:
                    for i in idx[0]:
                        new_var[i].domain = v[0][i].domain
                        new_var[i].setlb(v[0][i].lb)
                        new_var[i].setub(v[0][i].ub)
                c.vars[k] = new_var

        return ref



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import itertools
from pyomo.core.base.var import Var
from pyomo.core.base.constraint import Constraint
from pyomo.core.base.objective import Objective
from pyomo.core.expr.visitor import identify_variables
from pyomo.common.timing import HierarchicalTimer
from pyomo.util.subsystems import create_subsystem_block
from pyomo.contrib.pynumero.interfaces.pyomo_nlp import PyomoNLP
from pyomo.contrib.pynumero.interfaces.external_grey_box import ExternalGreyBoxModel
from pyomo.contrib.pynumero.algorithms.solvers.implicit_functions import (
    SccImplicitFunctionSolver,
)
import numpy as np
import scipy.sparse as sps


def _dense_to_full_sparse(matrix):
    """
    Used to convert a dense matrix (2d NumPy array) to SciPy sparse matrix
    with explicit coordinates for every entry, including zeros. This is
    used because _ExternalGreyBoxAsNLP methods rely on receiving sparse
    matrices where sparsity structure does not change between calls.
    This is difficult to achieve for matrices obtained via the implicit
    function theorem unless an entry is returned for every coordinate
    of the matrix.

    Note that this does not mean that the Hessian of the entire NLP will
    be dense, only that the block corresponding to this external model
    will be dense.
    """
    # TODO: Allow methods to hard-code Jacobian/Hessian sparsity structure
    # in the case it is known a priori.
    # TODO: Decompose matrices to infer maximum fill-in sparsity structure.
    nrow, ncol = matrix.shape
    row = []
    col = []
    data = []
    for i, j in itertools.product(range(nrow), range(ncol)):
        row.append(i)
        col.append(j)
        data.append(matrix[i, j])
    row = np.array(row)
    col = np.array(col)
    data = np.array(data)
    return sps.coo_matrix((data, (row, col)), shape=(nrow, ncol))


def get_hessian_of_constraint(constraint, wrt1=None, wrt2=None, nlp=None):
    constraints = [constraint]
    if wrt1 is None and wrt2 is None:
        variables = list(identify_variables(constraint.expr, include_fixed=False))
        wrt1 = variables
        wrt2 = variables
    elif wrt1 is not None and wrt2 is not None:
        variables = wrt1 + wrt2
    elif wrt1 is not None:  # but wrt2 is None
        wrt2 = wrt1
        variables = wrt1
    else:
        # wrt2 is not None and wrt1 is None
        wrt1 = wrt2
        variables = wrt1

    if nlp is None:
        block = create_subsystem_block(constraints, variables=variables)
        # Could fix input_vars so I don't evaluate the Hessian with respect
        # to variables I don't care about...

        # HUGE HACK: Variables not included in a constraint are not written
        # to the nl file, so we cannot take the derivative with respect to
        # them, even though we know this derivative is zero. To work around,
        # we make sure all variables appear on the block in the form of a
        # dummy constraint. Then we can take derivatives of any constraint
        # with respect to them. Conveniently, the extract_submatrix_
        # call deals with extracting the variables and constraint we care
        # about, in the proper order.
        block._dummy_var = Var()
        block._dummy_con = Constraint(expr=sum(variables) == block._dummy_var)
        block._obj = Objective(expr=0.0)
        nlp = PyomoNLP(block)

    saved_duals = nlp.get_duals()
    saved_obj_factor = nlp.get_obj_factor()
    temp_duals = np.zeros(len(saved_duals))

    # NOTE: This makes some assumption about how the Lagrangian is constructed.
    # TODO: Define the convention we assume and convert if necessary.
    idx = nlp.get_constraint_indices(constraints)[0]
    temp_duals[idx] = 1.0
    nlp.set_duals(temp_duals)
    nlp.set_obj_factor(0.0)

    # NOTE: The returned matrix preserves explicit zeros. I.e. it contains
    # coordinates for every entry that could possibly be nonzero.
    submatrix = nlp.extract_submatrix_hessian_lag(wrt1, wrt2)

    nlp.set_obj_factor(saved_obj_factor)
    nlp.set_duals(saved_duals)
    return submatrix


class ExternalPyomoModel(ExternalGreyBoxModel):
    """
    This is an ExternalGreyBoxModel used to create an external model
    from existing Pyomo components. Given a system of variables and
    equations partitioned into "input" and "external" variables and
    "residual" and "external" equations, this class computes the
    residual of the "residual equations," as well as their Jacobian
    and Hessian, as a function of only the inputs.

    Pyomo components:
        f(x, y) == 0 # "Residual equations"
        g(x, y) == 0 # "External equations", dim(g) == dim(y)

    Effective constraint seen by this "external model":
        F(x) == f(x, y(x)) == 0
        where y(x) solves g(x, y) == 0

    """

    def __init__(
        self,
        input_vars,
        external_vars,
        residual_cons,
        external_cons,
        solver_class=None,
        solver_options=None,
        timer=None,
    ):
        """
        Arguments:
        ----------
        input_vars: list
            List of variables sent to this system by the outer solver
        external_vars: list
            List of variables that are solved for internally by this system
        residual_cons: list
            List of equality constraints whose residuals are exposed to
            the outer solver
        external_cons: list
            List of equality constraints used to solve for the external
            variables
        solver_class: Subclass of ImplicitFunctionSolver
            The solver object that is used to converge the system of
            equations defining the implicit function.
        solver_options: dict
            Options dict for the ImplicitFunctionSolver
        timer: HierarchicalTimer
            HierarchicalTimer object to which new timing categories introduced
            will be attached. If None, a new timer will be created.

        """
        if timer is None:
            timer = HierarchicalTimer()
        self._timer = timer
        if solver_class is None:
            solver_class = SccImplicitFunctionSolver
        self._solver_class = solver_class
        if solver_options is None:
            solver_options = {}

        self._timer.start("__init__")

        # We only need this block to construct the NLP, which wouldn't
        # be necessary if we could compute Hessians of Pyomo constraints.
        self._block = create_subsystem_block(
            residual_cons + external_cons, input_vars + external_vars
        )
        self._block._obj = Objective(expr=0.0)
        self._timer.start("PyomoNLP")
        self._nlp = PyomoNLP(self._block)
        self._timer.stop("PyomoNLP")

        # Instantiate a solver with the ImplicitFunctionSolver API:
        self._solver = self._solver_class(
            external_vars,
            external_cons,
            input_vars,
            timer=self._timer,
            **solver_options,
        )

        assert len(external_vars) == len(external_cons)

        self.input_vars = input_vars
        self.external_vars = external_vars
        self.residual_cons = residual_cons
        self.external_cons = external_cons

        self.residual_con_multipliers = [None for _ in residual_cons]
        self.residual_scaling_factors = None

        self._input_output_coords = self._nlp.get_primal_indices(
            input_vars + external_vars
        )

        self._timer.stop("__init__")

    def n_inputs(self):
        return len(self.input_vars)

    def n_equality_constraints(self):
        return len(self.residual_cons)

    # I would like to try to get by without using the following "name" methods.
    def input_names(self):
        return ["input_%i" % i for i in range(self.n_inputs())]

    def equality_constraint_names(self):
        return ["residual_%i" % i for i in range(self.n_equality_constraints())]

    def set_input_values(self, input_values):
        self._timer.start("set_inputs")

        solver = self._solver
        external_cons = self.external_cons
        external_vars = self.external_vars
        input_vars = self.input_vars

        solver.set_parameters(input_values)
        outputs = solver.evaluate_outputs()
        solver.update_pyomo_model()

        #
        # Send updated variable values to NLP for dervative evaluation
        #
        primals = self._nlp.get_primals()
        values = np.concatenate((input_values, outputs))
        primals[self._input_output_coords] = values
        self._nlp.set_primals(primals)

        self._timer.stop("set_inputs")

    def set_equality_constraint_multipliers(self, eq_con_multipliers):
        """
        Sets multipliers for residual equality constraints seen by the
        outer solver.

        """
        for i, val in enumerate(eq_con_multipliers):
            self.residual_con_multipliers[i] = val

    def set_external_constraint_multipliers(self, eq_con_multipliers):
        eq_con_multipliers = np.array(eq_con_multipliers)
        external_multipliers = self.calculate_external_constraint_multipliers(
            eq_con_multipliers
        )
        multipliers = np.concatenate((eq_con_multipliers, external_multipliers))
        cons = self.residual_cons + self.external_cons
        n_con = len(cons)
        assert n_con == self._nlp.n_constraints()
        duals = np.zeros(n_con)
        indices = self._nlp.get_constraint_indices(cons)
        duals[indices] = multipliers
        self._nlp.set_duals(duals)

    def calculate_external_constraint_multipliers(self, resid_multipliers):
        """
        Calculates the multipliers of the external constraints from the
        multipliers of the residual constraints (which are provided by
        the "outer" solver).

        """
        # NOTE: This method implicitly relies on the value of inputs stored
        # in the nlp. Should we also rely on the multiplier that are in
        # the nlp?
        # We would then need to call nlp.set_duals twice. Once with the
        # residual multipliers and once with the full multipliers.
        # I like the current approach better for now.
        nlp = self._nlp
        y = self.external_vars
        f = self.residual_cons
        g = self.external_cons
        jfy = nlp.extract_submatrix_jacobian(y, f)
        jgy = nlp.extract_submatrix_jacobian(y, g)

        jgy_t = jgy.transpose()
        jfy_t = jfy.transpose()
        dfdg = -sps.linalg.splu(jgy_t.tocsc()).solve(jfy_t.toarray())
        resid_multipliers = np.array(resid_multipliers)
        external_multipliers = dfdg.dot(resid_multipliers)
        return external_multipliers

    def get_full_space_lagrangian_hessians(self):
        """
        Calculates terms of Hessian of full-space Lagrangian due to
        external and residual constraints. Note that multipliers are
        set by set_equality_constraint_multipliers. These matrices
        are used to calculate the Hessian of the reduced-space
        Lagrangian.

        """
        nlp = self._nlp
        x = self.input_vars
        y = self.external_vars
        hlxx = nlp.extract_submatrix_hessian_lag(x, x)
        hlxy = nlp.extract_submatrix_hessian_lag(x, y)
        hlyy = nlp.extract_submatrix_hessian_lag(y, y)
        return hlxx, hlxy, hlyy

    def calculate_reduced_hessian_lagrangian(self, hlxx, hlxy, hlyy):
        """
        Performs the matrix multiplications necessary to get the
        reduced space Hessian-of-Lagrangian term from the full-space
        terms.

        """
        # Converting to dense is faster for the distillation
        # example. Does this make sense?
        hlxx = hlxx.toarray()
        hlxy = hlxy.toarray()
        hlyy = hlyy.toarray()
        dydx = self.evaluate_jacobian_external_variables()
        term1 = hlxx
        prod = hlxy.dot(dydx)
        term2 = prod + prod.transpose()
        term3 = hlyy.dot(dydx).transpose().dot(dydx)
        hess_lag = term1 + term2 + term3
        return hess_lag

    def evaluate_equality_constraints(self):
        return self._nlp.extract_subvector_constraints(self.residual_cons)

    def evaluate_jacobian_equality_constraints(self):
        self._timer.start("jacobian")

        nlp = self._nlp
        x = self.input_vars
        y = self.external_vars
        f = self.residual_cons
        g = self.external_cons
        jfx = nlp.extract_submatrix_jacobian(x, f)
        jfy = nlp.extract_submatrix_jacobian(y, f)
        jgx = nlp.extract_submatrix_jacobian(x, g)
        jgy = nlp.extract_submatrix_jacobian(y, g)

        nf = len(f)
        nx = len(x)
        n_entries = nf * nx

        # TODO: Does it make sense to cast dydx to a sparse matrix?
        # My intuition is that it does only if jgy is "decomposable"
        # in the strongly connected component sense, which is probably
        # not usually the case.
        dydx = -1 * sps.linalg.splu(jgy.tocsc()).solve(jgx.toarray())
        # NOTE: PyNumero block matrices require this to be a sparse matrix
        # that contains coordinates for every entry that could possibly
        # be nonzero. Here, this is all of the entries.
        dfdx = jfx + jfy.dot(dydx)

        full_sparse = _dense_to_full_sparse(dfdx)

        self._timer.stop("jacobian")
        return full_sparse

    def evaluate_jacobian_external_variables(self):
        nlp = self._nlp
        x = self.input_vars
        y = self.external_vars
        g = self.external_cons
        jgx = nlp.extract_submatrix_jacobian(x, g)
        jgy = nlp.extract_submatrix_jacobian(y, g)
        jgy_csc = jgy.tocsc()
        dydx = -1 * sps.linalg.splu(jgy_csc).solve(jgx.toarray())
        return dydx

    def evaluate_hessian_external_variables(self):
        nlp = self._nlp
        x = self.input_vars
        y = self.external_vars
        g = self.external_cons
        jgx = nlp.extract_submatrix_jacobian(x, g)
        jgy = nlp.extract_submatrix_jacobian(y, g)
        jgy_csc = jgy.tocsc()
        jgy_fact = sps.linalg.splu(jgy_csc)
        dydx = -1 * jgy_fact.solve(jgx.toarray())

        ny = len(y)
        nx = len(x)

        hgxx = np.array(
            [get_hessian_of_constraint(con, x, nlp=nlp).toarray() for con in g]
        )
        hgxy = np.array(
            [get_hessian_of_constraint(con, x, y, nlp=nlp).toarray() for con in g]
        )
        hgyy = np.array(
            [get_hessian_of_constraint(con, y, nlp=nlp).toarray() for con in g]
        )

        # This term is sparse, but we do not exploit it.
        term1 = hgxx

        # This is what we want.
        # prod[i,j,k] = sum(hgxy[i,:,j] * dydx[:,k])
        prod = hgxy.dot(dydx)
        # Swap the second and third axes of the tensor
        term2 = prod + prod.transpose((0, 2, 1))
        # The term2 tensor could have some sparsity worth exploiting.

        # matrix.dot(tensor) is not what we want, so we reverse the order of the
        # product. Exploit symmetry of hgyy to only perform one transpose.
        term3 = hgyy.dot(dydx).transpose((0, 2, 1)).dot(dydx)

        rhs = term1 + term2 + term3

        rhs.shape = (ny, nx * nx)
        sol = jgy_fact.solve(rhs)
        sol.shape = (ny, nx, nx)
        d2ydx2 = -sol

        return d2ydx2

    def evaluate_hessians_of_residuals(self):
        """
        This method computes the Hessian matrix of each equality
        constraint individually, rather than the sum of Hessians
        times multipliers.
        """
        nlp = self._nlp
        x = self.input_vars
        y = self.external_vars
        f = self.residual_cons
        g = self.external_cons
        jfx = nlp.extract_submatrix_jacobian(x, f)
        jfy = nlp.extract_submatrix_jacobian(y, f)

        dydx = self.evaluate_jacobian_external_variables()

        ny = len(y)
        nf = len(f)
        nx = len(x)

        hfxx = np.array(
            [get_hessian_of_constraint(con, x, nlp=nlp).toarray() for con in f]
        )
        hfxy = np.array(
            [get_hessian_of_constraint(con, x, y, nlp=nlp).toarray() for con in f]
        )
        hfyy = np.array(
            [get_hessian_of_constraint(con, y, nlp=nlp).toarray() for con in f]
        )

        d2ydx2 = self.evaluate_hessian_external_variables()

        term1 = hfxx
        prod = hfxy.dot(dydx)
        term2 = prod + prod.transpose((0, 2, 1))
        term3 = hfyy.dot(dydx).transpose((0, 2, 1)).dot(dydx)

        d2ydx2.shape = (ny, nx * nx)
        term4 = jfy.dot(d2ydx2)
        term4.shape = (nf, nx, nx)

        d2fdx2 = term1 + term2 + term3 + term4
        return d2fdx2

    def evaluate_hessian_equality_constraints(self):
        """
        This method actually evaluates the sum of Hessians times
        multipliers, i.e. the term in the Hessian of the Lagrangian
        due to these equality constraints.

        """
        self._timer.start("hessian")

        # External multipliers must be calculated after both primals and duals
        # are set, and are only necessary for this Hessian calculation.
        # We know this Hessian calculation wants to use the most recently
        # set primals and duals, so we can safely calculate external
        # multipliers here.
        eq_con_multipliers = self.residual_con_multipliers
        self.set_external_constraint_multipliers(eq_con_multipliers)

        # These are full-space Hessian-of-Lagrangian terms
        hlxx, hlxy, hlyy = self.get_full_space_lagrangian_hessians()

        # These terms can be used to calculate the corresponding
        # Hessian-of-Lagrangian term in the full space.
        hess_lag = self.calculate_reduced_hessian_lagrangian(hlxx, hlxy, hlyy)
        sparse = _dense_to_full_sparse(hess_lag)
        lower_triangle = sps.tril(sparse)
        self._timer.stop("hessian")
        return lower_triangle

    def set_equality_constraint_scaling_factors(self, scaling_factors):
        """
        Set scaling factors for the equality constraints that are exposed
        to a solver. These are the "residual equations" in this class.
        """
        self.residual_scaling_factors = np.array(scaling_factors)

    def get_equality_constraint_scaling_factors(self):
        """
        Get scaling factors for the equality constraints that are exposed
        to a solver. These are the "residual equations" in this class.
        """
        return self.residual_scaling_factors



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common import Factory

WriterFactory = Factory('problem writer')


class AbstractProblemWriter:
    """Base class that can write optimization problems."""

    def __init__(self, problem_format):  # pragma:nocover
        self.format = problem_format

    def __call__(self, model, filename, solver_capability, **kwds):  # pragma:nocover
        raise TypeError(
            "Method __call__ undefined in writer for format " + str(self.format)
        )

    #
    # Support "with" statements.
    #
    def __enter__(self):
        return self

    def __exit__(self, t, v, traceback):
        pass


class BranchDirection:
    """Allowed values for MIP variable branching directions in the `direction` Suffix of a model."""

    default = 0
    down = -1
    up = 1

    ALL = {default, down, up}



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""
Between Steps (P-Split) reformulation for GDPs from [KMT21]_.

"""

from pyomo.common.config import (
    ConfigBlock,
    ConfigValue,
    document_kwargs_from_configdict,
)
from pyomo.common.modeling import unique_component_name
from pyomo.core import (
    Block,
    Constraint,
    Var,
    SortComponents,
    Transformation,
    TransformationFactory,
    TraversalStrategy,
    NonNegativeIntegers,
    value,
    ConcreteModel,
    Objective,
    ComponentMap,
    BooleanVar,
    LogicalConstraint,
    Connector,
    Expression,
    Suffix,
    Param,
    Set,
    SetOf,
    RangeSet,
    Reference,
    Binary,
    LogicalConstraintList,
    maximize,
)
from pyomo.core.base.external import ExternalFunction
from pyomo.network import Port
from pyomo.common.collections import ComponentSet
from pyomo.repn import generate_standard_repn
import pyomo.core.expr as EXPR
from pyomo.opt import SolverFactory
from pyomo.util.vars_from_expressions import get_vars_from_components

from pyomo.gdp import Disjunct, Disjunction, GDP_Error
from pyomo.gdp.util import (
    is_child_of,
    _to_dict,
    verify_successful_solve,
    NORMAL,
    clone_without_expression_components,
    _warn_for_active_disjunct,
    get_gdp_tree,
)
from pyomo.core.util import target_list
from pyomo.contrib.fbbt.fbbt import compute_bounds_on_expr
from weakref import ref as weakref_ref

from math import floor

import logging

logger = logging.getLogger('pyomo.gdp.partition_disjuncts')


def _generate_additively_separable_repn(nonlinear_part):
    if nonlinear_part.__class__ is not EXPR.SumExpression:
        # This isn't separable, so we just have the one expression
        return {
            'nonlinear_vars': [
                tuple(v for v in EXPR.identify_variables(nonlinear_part))
            ],
            'nonlinear_exprs': [nonlinear_part],
        }

    # else, it was a SumExpression, and we will break it into the summands,
    # recording which variables are there.
    nonlinear_decomp = {'nonlinear_vars': [], 'nonlinear_exprs': []}
    for summand in nonlinear_part.args:
        nonlinear_decomp['nonlinear_exprs'].append(summand)
        nonlinear_decomp['nonlinear_vars'].append(
            tuple(v for v in EXPR.identify_variables(summand))
        )

    return nonlinear_decomp


def arbitrary_partition(disjunction, P):
    """Returns a valid partition into P sets of the variables that appear in
    algebraic additively separable constraints in the Disjuncts in
    'disjunction'. Note that this method may return an invalid partition
    if the constraints are not additively separable!

    Arguments:
    ----------
    disjunction : DisjunctionData
        A Disjunction object for which the variable partition will be
        created.

    P : int
        the number of partitions

    """
    # collect variables
    v_set = ComponentSet()
    for disj in disjunction.disjuncts:
        v_set.update(
            get_vars_from_components(disj, Constraint, descend_into=Block, active=True)
        )
    # assign them to partitions
    partitions = [ComponentSet() for i in range(P)]
    for i, v in enumerate(v_set):
        partitions[i % P].add(v)

    return partitions


def compute_optimal_bounds(expr, global_constraints, opt):
    """Returns a tuple (LB, UB) where LB and UB are the results of minimizing
    and maximizing expr over the variable bounds and the constraints on the
    global_constraints block. Note that if expr is nonlinear, even if one of
    the min and max problems is convex, the other won't be!

    Arguments:
    ----------
    expr : ExpressionBase
        The subexpression whose bounds we will return

    global_constraints : BlockData
        A Block which contains the global Constraints and Vars of the
        original model

    opt : SolverBase
         A configured Solver object to use to minimize and maximize expr
         over the set defined by global_constraints. Note that if expr
         is nonlinear, opt will need to be capable of optimizing
         nonconvex problems.

    """
    if opt is None:
        raise GDP_Error(
            "No solver was specified to optimize the "
            "subproblems for computing expression bounds! "
            "Please specify a configured solver in the "
            "'compute_bounds_solver' argument if using "
            "'compute_optimal_bounds.'"
        )

    # add temporary objective and calculate bounds
    obj = Objective(expr=expr)
    global_constraints.add_component(
        unique_component_name(global_constraints, "tmp_obj"), obj
    )
    # Solve first minimizing, to get a lower bound
    results = opt.solve(global_constraints)
    if verify_successful_solve(results) is not NORMAL:
        logger.warning(
            "Problem to find lower bound for expression %s"
            "did not solve normally.\n\n%s" % (expr, results)
        )
        LB = None
    else:
        # This has some risks, if you're using a solver the gives a lower bound,
        # getting that would be better. But this is why this is a callback.
        LB = value(obj.expr)
    # Now solve maximizing, to get an upper bound
    obj.sense = maximize
    results = opt.solve(global_constraints)
    if verify_successful_solve(results) is not NORMAL:
        logger.warning(
            "Problem to find upper bound for expression %s"
            "did not solve normally.\n\n%s" % (expr, results)
        )
        UB = None
    else:
        UB = value(obj.expr)

    # clean up
    global_constraints.del_component(obj)
    del obj

    return (LB, UB)


def compute_fbbt_bounds(expr, global_constraints, opt):
    """
    Calls fbbt on expr and returns the lower and upper bounds on the expression
    based on the bounds of the Vars that appear in the expression. Ignores
    the global_constraints and opt arguments.
    """
    return compute_bounds_on_expr(expr)


@TransformationFactory.register(
    'gdp.partition_disjuncts',
    doc="Reformulates a convex disjunctive model "
    "into a new GDP by splitting additively "
    "separable constraints on P sets of variables",
)
@document_kwargs_from_configdict('CONFIG')
class PartitionDisjuncts_Transformation(Transformation):
    """
    Transform disjunctive model to equivalent disjunctive model (with
    potentially tighter hull relaxation) by taking the "P-split" formulation
    from Kronqvist et al. 2021 [KMT21]_. In each Disjunct, convex and additively
    separable constraints are split into separate constraints by introducing
    auxiliary variables that upperbound the subexpressions created by the split.
    Increasing the number of partitions can result in tighter hull relaxations,
    but at the cost of larger model sizes.

    The transformation will create a new Block with a unique name beginning
    "_pyomo_gdp_partition_disjuncts_reformulation".
    The Block will have new Disjunct objects, each corresponding to one of the
    Disjuncts being transformed. These will have the transformed constraints on
    them, and be in new Disjunctions, each corresponding to one of the
    originals. In addition, the auxiliary variables and the partitioned
    constraints will be declared on this Block, as well as LogicalConstraints
    linking the original indicator_vars with the ones of the transformed
    Disjuncts. All original GDP components that were transformed will be
    deactivated.

    References
    ----------
    See [KMT21]_.

    """

    CONFIG = ConfigBlock("gdp.partition_disjuncts")
    CONFIG.declare(
        'targets',
        ConfigValue(
            default=None,
            domain=target_list,
            description="""target or list of targets that will be relaxed""",
            doc="""
        Specifies the target or list of targets to relax as either a
        component or a list of components. 

        If None (default), the entire model is transformed. Note that if the 
        transformation is done out of place, the list of targets should be 
        attached to the model before it is cloned, and the list will specify 
        the targets on the cloned instance.
        """,
        ),
    )
    CONFIG.declare(
        'variable_partitions',
        ConfigValue(
            default=None,
            domain=_to_dict,
            description="""Set of sets of variables which define valid partitions
        (i.e., the constraints are additively separable across these
        partitions). These can be specified globally (for all active
        Disjunctions), or by Disjunction.""",
            doc="""
        Specified variable partitions, either globally or per Disjunction.

        Expects either a set of disjoint ComponentSets whose union is all the
        variables that appear in all Disjunctions or a mapping from each active
        Disjunction to a set of disjoint ComponentSets whose union is the set
        of variables that appear in that Disjunction. In either case, if any
        constraints in the Disjunction are only partially additively separable,
        these sets must be a valid partition so that these constraints are
        additively separable with respect to this partition. To specify a
        default partition for Disjunctions that do not appear as keys in the
        map, map the partition to 'None.'

        Last, note that in the case of constraints containing partially
        additively separable functions, it is required that the user specify
        the variable partition(s).
        """,
        ),
    )
    CONFIG.declare(
        'num_partitions',
        ConfigValue(
            default=None,
            domain=_to_dict,
            description="""Number of partitions of variables, if variable_partitions
        is not specified. Can be specified separately for specific Disjunctions
        if desired.""",
            doc="""
        Either a single value so that all Disjunctions will have variables
        partitioned into P sets, or a map of Disjunctions to a value of P
        for each active Disjunction. Mapping None to a value of P will specify
        the default value of P to use if the value for a given Disjunction
        is not explicitly specified.

        Note that if any constraints contain partially additively separable
        functions, the partitions for the Disjunctions with these Constraints
        must be specified in the variable_partitions argument.
        """,
        ),
    )
    CONFIG.declare(
        'variable_partitioning_method',
        ConfigValue(
            default=arbitrary_partition,
            domain=_to_dict,
            description="""Method to partition the variables. By default, the
        partitioning will be done arbitrarily.""",
            doc="""
        A function which takes a Disjunction object and a number P and return
        a valid partitioning of the variables that appear in the disjunction
        into P partitions.

        Note that you must give a value for 'P' if you are using this method
        to calculate partitions.

        Note that if any constraints contain partially additively separable
        functions, the partitions for the Disjunctions cannot be calculated
        automatically. Please specify the partitions for the Disjunctions with
        these Constraints in the variable_partitions argument.
        """,
        ),
    )
    CONFIG.declare(
        'assume_fixed_vars_permanent',
        ConfigValue(
            default=False,
            domain=bool,
            description="""Boolean indicating whether or not to transform so that
        the transformed model will still be valid when fixed Vars are
        unfixed.""",
            doc="""
        If True, the transformation will create a correct model even if fixed
        variables are later unfixed. That is, bounds will be calculated based
        on fixed variables' bounds, not their values. However, if fixed
        variables will never be unfixed, a possibly tighter model will result,
        and fixed variables need not have bounds.

        Note that this has no effect on fixed BooleanVars, including the
        indicator variables of Disjuncts. The transformation is always correct
        whether or not these remain fixed.
        """,
        ),
    )
    CONFIG.declare(
        'compute_bounds_method',
        ConfigValue(
            default=compute_fbbt_bounds,
            description="""Function that takes an expression, a Block containing
        the global constraints of the original problem, and a configured
        solver, and returns both a lower and upper bound for the expression.""",
            doc="""
        Callback for computing bounds on expressions, in order to bound
        the auxiliary variables created by the transformation. 

        Some pre-implemented options include

           * compute_fbbt_bounds (the default), and
           * compute_optimal_bounds

        or you can write your own callback which accepts an Expression object,
        a model containing the variables and global constraints of the original
        instance, and a configured solver and returns a tuple (LB, UB) where
        either element can be None if no valid bound could be found.
        """,
        ),
    )
    CONFIG.declare(
        'compute_bounds_solver',
        ConfigValue(
            default=None,
            description="""Solver object to pass to compute_bounds_method.
        This is required if you are using 'compute_optimal_bounds'.""",
            doc="""
        Configured solver object for use in the compute_bounds_method.

        In particular, if compute_bounds_method is 'compute_optimal_bounds',
        this will be used to solve the subproblems, so needs to handle
        non-convex problems if any Disjunctions contain nonlinear constraints.
        """,
        ),
    )

    def __init__(self):
        super(PartitionDisjuncts_Transformation, self).__init__()
        self.handlers = {
            Constraint: self._transform_constraint,
            Var: False,  # these will be already dealt with--we add
            # references to them before we call handlers.
            BooleanVar: False,
            Connector: False,
            Expression: False,
            Suffix: False,
            Param: False,
            Set: False,
            SetOf: False,
            RangeSet: False,
            Disjunct: self._warn_for_active_disjunct,
            Block: False,
            ExternalFunction: False,
            Port: False,  # not Arcs, because those are deactivated after
            # the network.expand_arcs transformation
        }

    def _apply_to(self, instance, **kwds):
        if not instance.ctype in (Block, Disjunct):
            raise GDP_Error(
                "Transformation called on %s of type %s. 'instance'"
                " must be a ConcreteModel, Block, or Disjunct (in "
                "the case of nested disjunctions)." % (instance.name, instance.ctype)
            )
        try:
            self._config = self.CONFIG(kwds.pop('options', {}))
            self._config.set_value(kwds)
            self._transformation_blocks = {}

            if not self._config.assume_fixed_vars_permanent:
                fixed_vars = ComponentMap()
                for v in get_vars_from_components(
                    instance,
                    Constraint,
                    include_fixed=True,
                    active=True,
                    descend_into=(Block, Disjunct),
                ):
                    if v.fixed:
                        fixed_vars[v] = value(v)
                        v.fixed = False

            self._apply_to_impl(instance)

        finally:
            # restore fixed variables
            if not self._config.assume_fixed_vars_permanent:
                for v, val in fixed_vars.items():
                    v.fix(val)

            del self._config
            del self._transformation_blocks

    def _apply_to_impl(self, instance):
        self.variable_partitions = (
            self._config.variable_partitions
            if self._config.variable_partitions is not None
            else {}
        )
        self.partitioning_method = self._config.variable_partitioning_method

        # create a model to store the global constraints on that we will pass to
        # the compute_bounds_method, for if it wants them. We're making it a
        # separate model because we don't need it again
        global_constraints = ConcreteModel()
        for cons in instance.component_objects(
            Constraint,
            active=True,
            descend_into=Block,
            sort=SortComponents.deterministic,
        ):
            global_constraints.add_component(
                unique_component_name(
                    global_constraints, cons.getname(fully_qualified=True)
                ),
                Reference(cons),
            )
        for var in instance.component_objects(
            Var, descend_into=(Block, Disjunct), sort=SortComponents.deterministic
        ):
            global_constraints.add_component(
                unique_component_name(
                    global_constraints, var.getname(fully_qualified=True)
                ),
                Reference(var),
            )
        self._global_constraints = global_constraints

        # we can support targets as usual.
        targets = self._config.targets
        knownBlocks = {}
        if targets is None:
            targets = (instance,)
        # Disjunctions in targets will transform their Disjuncts which will in
        # turn transform all the GDP components declared on themselves. So we
        # only need to list root nodes of the GDP tree as targets, and
        # everything will be transformed (and in the correct order)
        targets = self._preprocess_targets(targets, instance, knownBlocks)
        for t in targets:
            if t.ctype is Disjunction:
                # After preprocessing, we know that this is not indexed.
                self._transform_disjunctionData(t, t.index())
            else:  # We know this is a DisjunctData after preprocessing
                self._transform_blockData(t)

    def _preprocess_targets(self, targets, instance, knownBlocks):
        gdp_tree = get_gdp_tree(targets, instance, knownBlocks)

        preprocessed_targets = []
        # We need only transform root nodes of the tree--the rest will be
        # transformed recursively from there. (It's also possible to do a
        # topological sort here and just make sure we don't ask for nested
        # Disjuncts after their Disjunctions, but that's more work than is
        # necessary.)
        for node in gdp_tree.vertices:
            if gdp_tree.in_degree(node) == 0:
                preprocessed_targets.append(node)

        return preprocessed_targets

    def _get_transformation_block(self, block):
        if self._transformation_blocks.get(block) is not None:
            return self._transformation_blocks[block]

        # create a transformation block on which we will create the reformulated
        # GDP...
        self._transformation_blocks[block] = transformation_block = Block()
        block.add_component(
            unique_component_name(
                block, '_pyomo_gdp_partition_disjuncts_reformulation'
            ),
            transformation_block,
        )

        transformation_block.indicator_var_equalities = LogicalConstraint(
            NonNegativeIntegers
        )

        return transformation_block

    def _transform_blockData(self, obj):
        # compute the list of Disjunctions to transform *once*, then do it. Else
        # we will pick up the Disjunctions we create!
        to_transform = []
        # Transform every (active) disjunction in the block. Don't descend into
        # Disjuncts because we'll transform what's on them recursively.
        for disjunction in obj.component_data_objects(
            Disjunction,
            active=True,
            sort=SortComponents.deterministic,
            descend_into=Block,
        ):
            to_transform.append(disjunction)

        for disjunction in to_transform:
            self._transform_disjunctionData(disjunction, disjunction.index())

    def _transform_disjunctionData(
        self, obj, idx, transBlock=None, transformed_parent_disjunct=None
    ):
        if not obj.active:
            return

        # Just because it's unlikely this is what someone meant to do...
        if len(obj.disjuncts) == 0:
            raise GDP_Error(
                "Disjunction '%s' is empty. This is "
                "likely indicative of a modeling error."
                % obj.getname(fully_qualified=True)
            )

        if transBlock is None and transformed_parent_disjunct is not None:
            transBlock = self._get_transformation_block(transformed_parent_disjunct)
        if transBlock is None:
            transBlock = self._get_transformation_block(obj.parent_block())

        variable_partitions = self.variable_partitions
        partition_method = self.partitioning_method

        # was the partition specified for the disjunct?
        partition = variable_partitions.get(obj)
        if partition is None:
            # was there a default partition?
            partition = variable_partitions.get(None)
            if partition is None:
                # If not, see what method to use to calculate one
                method = partition_method.get(obj)
                if method is None:
                    # was there a default method?
                    method = partition_method.get(None)
                # if all else fails, set it to our default
                method = method if method is not None else arbitrary_partition

                # now figure out P
                if self._config.num_partitions is None:
                    # This will just end in failure below. (We're checking here
                    # because we don't need a value of P if the partitions were
                    # specified for every Disjunction.)
                    P = None
                else:
                    P = self._config.num_partitions.get(obj)
                    if P is None:
                        P = self._config.num_partitions.get(None)
                if P is None:
                    raise GDP_Error(
                        "No value for P was given for disjunction "
                        "%s! Please specify a value of P "
                        "(number of "
                        "partitions), if you do not specify the "
                        "partitions directly." % obj.name
                    )
                # it's this method's job to scream if it can't handle what's
                # here, we can only assume it worked for now, since it's a
                # callback.
                partition = method(obj, P)
        # these have to be ComponentSets
        partition = [ComponentSet(var_list) for var_list in partition]

        transformed_disjuncts = []
        for disjunct in obj.disjuncts:
            transformed_disjunct = self._transform_disjunct(
                disjunct, partition, transBlock
            )
            if transformed_disjunct is not None:
                transformed_disjuncts.append(transformed_disjunct)
                # These require transformation, but that's okay because we are
                # going to a GDP
                transBlock.indicator_var_equalities[
                    len(transBlock.indicator_var_equalities)
                ] = disjunct.indicator_var.equivalent_to(
                    transformed_disjunct.indicator_var
                )

        # make a new disjunction with the transformed guys
        transformed_disjunction = Disjunction(
            expr=[disj for disj in transformed_disjuncts]
        )
        transBlock.add_component(
            unique_component_name(transBlock, obj.getname(fully_qualified=True)),
            transformed_disjunction,
        )
        obj._algebraic_constraint = weakref_ref(transformed_disjunction)

        obj.deactivate()

    def _get_leq_constraints(self, cons):
        constraints = []
        if cons.lower is not None:
            constraints.append((-cons.body, -cons.lower))
        if cons.upper is not None:
            constraints.append((cons.body, cons.upper))
        return constraints

    def _transform_disjunct(self, disjunct, partition, transBlock):
        # deactivated -> either we've already transformed or user deactivated
        if not disjunct.active:
            if disjunct.indicator_var.is_fixed():
                if not value(disjunct.indicator_var):
                    # The user cleanly deactivated the disjunct: there
                    # is nothing for us to do here.
                    return
                else:
                    raise GDP_Error(
                        "The disjunct '%s' is deactivated, but the "
                        "indicator_var is fixed to %s. This makes no sense."
                        % (disjunct.name, value(disjunct.indicator_var))
                    )
            if disjunct._transformation_block is None:
                raise GDP_Error(
                    "The disjunct '%s' is deactivated, but the "
                    "indicator_var is not fixed and the disjunct does not "
                    "appear to have been relaxed. This makes no sense. "
                    "(If the intent is to deactivate the disjunct, fix its "
                    "indicator_var to False.)" % (disjunct.name,)
                )

        if disjunct._transformation_block is not None:
            # we've transformed it, which means this is the second time it's
            # appearing in a Disjunction
            raise GDP_Error(
                "The disjunct '%s' has been transformed, but a disjunction "
                "it appears in has not. Putting the same disjunct in "
                "multiple disjunctions is not supported." % disjunct.name
            )

        transformed_disjunct = Disjunct()
        disjunct._transformation_block = weakref_ref(transformed_disjunct)
        transBlock.add_component(
            unique_component_name(transBlock, disjunct.getname(fully_qualified=True)),
            transformed_disjunct,
        )
        # If the original has an indicator_var fixed to something, fix this one
        # too.
        if disjunct.indicator_var.fixed:
            transformed_disjunct.indicator_var.fix(value(disjunct.indicator_var))

        # need to transform inner Disjunctions first (before we complain about
        # active Disjuncts)
        for disjunction in disjunct.component_data_objects(
            Disjunction,
            active=True,
            sort=SortComponents.deterministic,
            descend_into=Block,
        ):
            self._transform_disjunctionData(
                disjunction, disjunction.index(), None, transformed_disjunct
            )

        # create references to any variables declared here on the transformed
        # Disjunct (this will include the indicator_var) NOTE that we will not
        # have to do this when #1032 is implemented for the writers. But right
        # now, we are going to deactivate this and hide it from the active
        # subtree, so we need to be safe.
        for var in disjunct.component_objects(Var, descend_into=Block, active=None):
            transformed_disjunct.add_component(
                unique_component_name(
                    transformed_disjunct, var.getname(fully_qualified=True)
                ),
                Reference(var),
            )

        # Since this transformation is GDP -> GDP and it is based on
        # partitioning algebraic expressions, we will copy over
        # LogicalConstraints that may be on the Disjuncts, without transforming
        # them. This is consistent with our handling of nested Disjunctions,
        # which also remain nested, though their algebraic constraints may be
        # transformed. Note that we are not using References because when asked
        # who their parent block is, we would like these constraints to answer
        # that it is the transformed Disjunct.
        logical_constraints = LogicalConstraintList()
        transformed_disjunct.add_component(
            unique_component_name(transformed_disjunct, 'logical_constraints'),
            logical_constraints,
        )
        for cons in disjunct.component_data_objects(
            LogicalConstraint, descend_into=Block, active=None
        ):
            # Add a copy of it on the new Disjunct
            logical_constraints.add(cons.expr)

            # deactivate to mark as transformed (so we don't hit it in the loop
            # below)
            cons.deactivate()

        # transform everything else
        for obj in disjunct.component_data_objects(
            active=True, sort=SortComponents.deterministic, descend_into=Block
        ):
            handler = self.handlers.get(obj.ctype, None)
            if not handler:
                if handler is None:
                    raise GDP_Error(
                        "No partition_disjuncts transformation handler "
                        "registered "
                        "for modeling components of type %s. If your "
                        "disjuncts contain non-GDP Pyomo components that "
                        "require transformation, please transform them first."
                        % obj.ctype
                    )
                continue
            # we are really only transforming constraints and checking for
            # anything nutty (active Disjuncts, etc) here, so pass through what
            # is necessary for transforming Constraints
            handler(obj, disjunct, transformed_disjunct, transBlock, partition)

        disjunct._deactivate_without_fixing_indicator()
        return transformed_disjunct

    def _transform_constraint(
        self, cons, disjunct, transformed_disjunct, transBlock, partition
    ):
        instance = disjunct.model()
        cons_name = cons.getname(fully_qualified=True)

        # create place on transformed Disjunct for the new constraint and
        # for the auxiliary variables
        transformed_constraint = Constraint(NonNegativeIntegers)
        transformed_disjunct.add_component(
            unique_component_name(transformed_disjunct, cons_name),
            transformed_constraint,
        )
        aux_vars = Var(NonNegativeIntegers, dense=False)
        transformed_disjunct.add_component(
            unique_component_name(transformed_disjunct, cons_name + "_aux_vars"),
            aux_vars,
        )

        # create a place on the transBlock for the split constraints
        split_constraints = Constraint(NonNegativeIntegers)
        transBlock.add_component(
            unique_component_name(transBlock, cons_name + "_split_constraints"),
            split_constraints,
        )

        # this is a list which might have two constraints in it if we had
        # both a lower and upper value.
        leq_constraints = self._get_leq_constraints(cons)
        for body, rhs in leq_constraints:
            repn = generate_standard_repn(body, compute_values=True)
            nonlinear_repn = None
            if repn.nonlinear_expr is not None:
                nonlinear_repn = _generate_additively_separable_repn(
                    repn.nonlinear_expr
                )
            split_exprs = []
            split_aux_vars = []
            vars_not_accounted_for = ComponentSet(
                v for v in EXPR.identify_variables(body, include_fixed=False)
            )
            vars_accounted_for = ComponentSet()
            for idx, var_list in enumerate(partition):
                # we are going to recreate the piece of the expression
                # involving the vars in var_list
                split_exprs.append(0)
                expr = split_exprs[-1]
                for i, v in enumerate(repn.linear_vars):
                    if v in var_list:
                        expr += repn.linear_coefs[i] * v
                        vars_accounted_for.add(v)
                for i, (v1, v2) in enumerate(repn.quadratic_vars):
                    if v1 in var_list:
                        if v2 not in var_list:
                            raise GDP_Error(
                                "Variables '%s' and '%s' are "
                                "multiplied in Constraint '%s', "
                                "but they are in different "
                                "partitions! Please ensure that "
                                "all the constraints in the "
                                "disjunction are "
                                "additively separable with "
                                "respect to the specified "
                                "partition." % (v1.name, v2.name, cons.name)
                            )
                        expr += repn.quadratic_coefs[i] * v1 * v2
                        vars_accounted_for.add(v1)
                        vars_accounted_for.add(v2)
                if nonlinear_repn is not None:
                    for i, expr_var_set in enumerate(nonlinear_repn['nonlinear_vars']):
                        # check if v_list is a subset of var_list. If it is
                        # not and there is no intersection, we move on. If
                        # it is not and there is an intersection, we raise
                        # an error: It's not a valid partition. If it is,
                        # then we add this piece of the expression.
                        # subset?
                        if all(v in var_list for v in list(expr_var_set)):
                            expr += nonlinear_repn['nonlinear_exprs'][i]
                            for var in expr_var_set:
                                vars_accounted_for.add(var)
                        # intersection?
                        elif len(ComponentSet(expr_var_set) & var_list) != 0:
                            raise GDP_Error(
                                "Variables which appear in the "
                                "expression %s are in different "
                                "partitions, but this "
                                "expression doesn't appear "
                                "additively separable. Please "
                                "expand it if it is additively "
                                "separable or, more likely, "
                                "ensure that all the "
                                "constraints in the disjunction "
                                "are additively separable with "
                                "respect to the specified "
                                "partition. If you did not "
                                "specify a partition, only "
                                "a value of P, note that to "
                                "automatically partition the "
                                "variables, we assume all the "
                                "expressions are additively "
                                "separable." % nonlinear_repn['nonlinear_exprs'][i]
                            )

                expr_lb, expr_ub = self._config.compute_bounds_method(
                    expr, self._global_constraints, self._config.compute_bounds_solver
                )
                if expr_lb is None or expr_ub is None:
                    raise GDP_Error(
                        "Expression %s from constraint '%s' "
                        "is unbounded! Please ensure all "
                        "variables that appear "
                        "in the constraint are bounded or "
                        "specify compute_bounds_method="
                        "compute_optimal_bounds"
                        " if the expression is bounded by the "
                        "global constraints." % (expr, cons.name)
                    )
                # if the expression was empty wrt the partition, we don't
                # need to bother with any of this. The aux_var doesn't need
                # to exist because it would be 0.
                if type(expr) is not int or expr != 0:
                    aux_var = aux_vars[len(aux_vars)]
                    aux_var.setlb(expr_lb)
                    aux_var.setub(expr_ub)
                    split_aux_vars.append(aux_var)
                    split_constraints[len(split_constraints)] = expr <= aux_var

            if len(vars_accounted_for) < len(vars_not_accounted_for):
                orphans = vars_not_accounted_for - vars_accounted_for
                orphan_string = ""
                for v in orphans:
                    orphan_string += "'%s', " % v.name
                orphan_string = orphan_string[:-2]
                raise GDP_Error(
                    "Partition specified for disjunction "
                    "containing Disjunct '%s' does not "
                    "include all the variables that appear "
                    "in the disjunction. The following "
                    "variables are not assigned to any part "
                    "of the partition: %s" % (disjunct.name, orphan_string)
                )
            transformed_constraint[len(transformed_constraint)] = (
                sum(v for v in split_aux_vars) <= rhs - repn.constant
            )
        # deactivate the constraint since we've transformed it
        cons.deactivate()

    def _warn_for_active_disjunct(
        self,
        disjunct,
        parent_disjunct,
        transformed_parent_disjunct,
        transBlock,
        partition,
    ):
        _warn_for_active_disjunct(disjunct, parent_disjunct)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import bisect
import importlib
import pyomo.tpl.ply.lex as lex
import pyomo.tpl.ply.yacc as yacc

from pyomo.common.errors import DeveloperError
from pyomo.common.fileutils import this_file, this_file_dir
from pyomo.core.base.util import flatten_tuple

_re_number = r'[-+]?(?:[0-9]+\.?[0-9]*|\.[0-9]+)(?:[eE][-+]?[0-9]+)?'

## -----------------------------------------------------------
##
## Lexer definitions for tokenizing the input
##
## -----------------------------------------------------------

_parse_info = None

states = (('data', 'inclusive'),)

reserved = {
    'data': 'DATA',
    'set': 'SET',
    'param': 'PARAM',
    'end': 'END',
    'store': 'STORE',
    'load': 'LOAD',
    'table': 'TABLE',
    'include': 'INCLUDE',
    'namespace': 'NAMESPACE',
}

# Token names
tokens = [
    "COMMA",
    "LBRACE",
    "RBRACE",
    "SEMICOLON",
    "COLON",
    "COLONEQ",
    "LBRACKET",
    "RBRACKET",
    "LPAREN",
    "RPAREN",
    "WORD",
    "WORDWITHLBRACKET",
    "STRING",
    "BRACKETEDSTRING",
    "QUOTEDSTRING",
    "EQ",
    "TR",
    "ASTERISK",
    "NUM_VAL",
    # "NONWORD",
] + list(reserved.values())

# Ignore space and tab
t_ignore = " \t\r"

# Regular expression rules
t_COMMA = r","
t_LBRACKET = r"\["
t_RBRACKET = r"\]"
t_LBRACE = r"\{"
t_RBRACE = r"\}"
t_COLON = r":"
t_EQ = r"="
t_TR = r"\(tr\)"
t_LPAREN = r"\("
t_RPAREN = r"\)"
t_ASTERISK = r"\*"


#
# Notes on PLY tokenization
#   - token functions (beginning with "t_") are prioritized in the order
#     that they are declared in this module
#   - use @lex.TOKEN instead of docstrings to avoid errors from the
#     Sphinx autosummary
#
@lex.TOKEN(r'[\n]+')
def t_newline(t):
    t.lexer.lineno += len(t.value)
    t.lexer.linepos.extend(t.lexpos + i for i, _ in enumerate(t.value))


# Discard comments
_re_singleline_comment = r'(?:\#[^\n]*)'
_re_multiline_comment = r'(?:/\*(?:[\n]|.)*?\*/)'


@lex.TOKEN('|'.join([_re_singleline_comment, _re_multiline_comment]))
def t_COMMENT(t):
    # Single-line and multi-line strings
    nlines = t.value.count('\n')
    t.lexer.lineno += nlines
    # We will never need to determine column numbers within this comment
    # block, so it is sufficient to just worry about the *last* newline
    # in the comment
    lastpos = t.lexpos + t.value.rfind('\n')
    t.lexer.linepos.extend(lastpos for i in range(nlines))


@lex.TOKEN(r':=')
def t_COLONEQ(t):
    t.lexer.begin('data')
    return t


@lex.TOKEN(r';')
def t_SEMICOLON(t):
    t.lexer.begin('INITIAL')
    return t


# Numbers must be followed by a delimiter token (EOF is not a concern,
# as valid DAT files always end with a ';').
@lex.TOKEN(_re_number + r'(?=[\s()\[\]{}:;,])')
def t_NUM_VAL(t):
    _num = float(t.value)
    if '.' in t.value:
        t.value = _num
    else:
        _int = int(_num)
        t.value = _int if _num == _int else _num
    return t


@lex.TOKEN(r'[a-zA-Z_][a-zA-Z0-9_\.\-]*\[')
def t_WORDWITHLBRACKET(t):
    return t


@lex.TOKEN(r'[a-zA-Z_][a-zA-Z_0-9\.+\-]*')
def t_WORD(t):
    if t.value in reserved:
        t.type = reserved[t.value]  # Check for reserved words
    return t


@lex.TOKEN(r'[a-zA-Z0-9_\.+\-\\\/]+')
def t_STRING(t):
    # Note: RE guarantees the string has no embedded quotation characters
    t.value = '"' + t.value + '"'
    return t


@lex.TOKEN(r'[a-zA-Z0-9_\.+\-]*\[[a-zA-Z0-9_\.+\-\*,\s]+\]')
def t_data_BRACKETEDSTRING(t):
    # NO SPACES
    # a[1,_df,'foo bar']
    # [1,*,'foo bar']
    return t


_re_quoted_str = r'"(?:[^"]|"")*"'


@lex.TOKEN("|".join([_re_quoted_str, _re_quoted_str.replace('"', "'")]))
def t_QUOTEDSTRING(t):
    # Normalize the quotes to use '"', and replace doubled ("escaped")
    # quotation characters with a single character
    t.value = '"' + t.value[1:-1].replace(2 * t.value[0], t.value[0]) + '"'
    return t


# t_NONWORD   = r"[^\.A-Za-z0-9,;:=<>\*\(\)\#{}\[\] \n\t\r]+"


# Error handling rule
def t_error(t):
    raise IOError(
        "ERROR: Token %s Value %s Line %s Column %s"
        % (t.type, t.value, t.lineno, t.lexpos)
    )


## DEBUGGING: uncomment to get tokenization information
# def _wrap(_name, _fcn):
#     def _wrapper(t):
#         print(_name + ": %s" % (t.value,))
#         return _fcn(t)
#     _wrapper.__doc__ = _fcn.__doc__
#     return _wrapper
# import inspect
# for _name in list(globals()):
#     if _name.startswith('t_') and inspect.isfunction(globals()[_name]):
#         globals()[_name] = _wrap(_name, globals()[_name])


def _lex_token_position(t):
    i = bisect.bisect_left(t.lexer.linepos, t.lexpos)
    if i:
        return t.lexpos - t.lexer.linepos[i - 1]
    return t.lexpos


## -----------------------------------------------------------
##
## Yacc grammar for data commands
##
## -----------------------------------------------------------


def p_expr(p):
    '''expr : statements
    |'''
    if len(p) == 2:
        # print "STMTS",p[1]
        for stmt in p[1]:
            if type(stmt) is list:
                _parse_info[None].append(stmt)
            else:
                for key in stmt:
                    if key in _parse_info:
                        _parse_info[key].append(stmt[key])
                    else:
                        _parse_info[key] = stmt[key]


def p_statements(p):
    '''statements : statements statement
    | statement
    | statements NAMESPACE WORD LBRACE statements RBRACE
    | NAMESPACE WORD LBRACE statements RBRACE'''
    # print "STMT X",p[1:],p[1]
    len_p = len(p)
    if len_p == 3:
        # NB: statements will never be None, but statement *could* be None
        p[0] = p[1]
        if p[2] is not None:
            p[0].append(p[2])
    elif len_p == 2:
        if p[1] is None:
            p[0] = []
        else:
            p[0] = [p[1]]
    elif len_p == 7:
        # NB: statements will never be None
        p[0] = p[1]
        p[0].append({p[3]: p[5]})
    else:
        # NB: statements will never be None
        p[0] = [{p[2]: p[4]}]


def p_statement(p):
    '''statement : SET WORD COLONEQ datastar SEMICOLON
    | SET WORDWITHLBRACKET args RBRACKET COLONEQ datastar SEMICOLON
    | SET WORD COLON itemstar COLONEQ datastar SEMICOLON
    | PARAM items COLONEQ datastar SEMICOLON
    | TABLE items COLONEQ datastar SEMICOLON
    | LOAD items SEMICOLON
    | STORE items SEMICOLON
    | INCLUDE WORD SEMICOLON
    | INCLUDE QUOTEDSTRING SEMICOLON
    | DATA SEMICOLON
    | END SEMICOLON
    '''
    # print "STATEMENT",len(p), p[1:]
    stmt = p[1]
    if stmt == 'set':
        if p[2][-1] == '[':
            # Just turn off the flatten_list and see what happens
            p[0] = ['set', p[2][:-1], '['] + list(
                flatten_tuple([p[i] for i in range(3, len(p) - 1)])
            )
        else:
            p[0] = list(flatten_tuple([p[i] for i in range(1, len(p) - 1)]))
    elif stmt == 'param':
        p[0] = list(flatten_tuple([p[i] for i in range(1, len(p) - 1)]))
    elif stmt == 'include':
        p[0] = [p[i] for i in range(1, len(p) - 1)]
    elif stmt == 'load':
        p[0] = [p[1]] + p[2]
    elif stmt == 'store':
        p[0] = [p[1]] + p[2]
    elif stmt == 'table':
        p[0] = [p[1]] + [p[2]] + [p[4]]
    else:
        # Not necessary, but nice to document how statement could end up None
        p[0] = None
    # print(p[0])


def p_datastar(p):
    '''
    datastar : data
             |
    '''
    if len(p) == 2:
        p[0] = p[1]
    else:
        p[0] = []


def p_data(p):
    '''
    data : data NUM_VAL
         | data WORD
         | data STRING
         | data QUOTEDSTRING
         | data BRACKETEDSTRING
         | data SET
         | data TABLE
         | data PARAM
         | data LPAREN
         | data RPAREN
         | data COMMA
         | data ASTERISK
         | NUM_VAL
         | WORD
         | STRING
         | QUOTEDSTRING
         | BRACKETEDSTRING
         | SET
         | TABLE
         | PARAM
         | LPAREN
         | RPAREN
         | COMMA
         | ASTERISK
    '''
    # Locate and handle item as necessary
    single_item = len(p) == 2
    if single_item:
        tmp = p[1]
    else:
        tmp = p[2]
    # if type(tmp) is str and tmp[0] == '"' and tmp[-1] == '"' and len(tmp) > 2 and not ' ' in tmp:
    #    tmp = tmp[1:-1]

    # Grow items list according to parsed item length
    if single_item:
        p[0] = [tmp]
    else:
        # yacc __getitem__ is expensive: use a local list to avoid a
        # getitem call on p[0]
        tmp_lst = p[1]
        tmp_lst.append(tmp)
        p[0] = tmp_lst


def p_args(p):
    '''
    args : arg
         |
    '''
    if len(p) == 2:
        p[0] = p[1]
    else:
        p[0] = []


def p_arg(p):
    '''
    arg : arg COMMA NUM_VAL
         | arg COMMA WORD
         | arg COMMA STRING
         | arg COMMA QUOTEDSTRING
         | arg COMMA SET
         | arg COMMA TABLE
         | arg COMMA PARAM
         | NUM_VAL
         | WORD
         | STRING
         | QUOTEDSTRING
         | SET
         | TABLE
         | PARAM
    '''
    # Locate and handle item as necessary
    single_item = len(p) == 2
    if single_item:
        tmp = p[1]
    else:
        tmp = p[3]
    if (
        type(tmp) is str
        and tmp[0] == '"'
        and tmp[-1] == '"'
        and len(tmp) > 2
        and not ' ' in tmp
    ):
        tmp = tmp[1:-1]

    # Grow items list according to parsed item length
    if single_item:
        p[0] = [tmp]
    else:
        # yacc __getitem__ is expensive: use a local list to avoid a
        # getitem call on p[0]
        tmp_lst = p[1]
        tmp_lst.append(tmp)
        p[0] = tmp_lst


def p_itemstar(p):
    '''
    itemstar : items
             |
    '''
    if len(p) == 2:
        p[0] = p[1]
    else:
        p[0] = []


def p_items(p):
    '''
    items : items NUM_VAL
          | items WORD
          | items STRING
          | items QUOTEDSTRING
          | items COMMA
          | items COLON
          | items LBRACE
          | items RBRACE
          | items LBRACKET
          | items RBRACKET
          | items TR
          | items LPAREN
          | items RPAREN
          | items ASTERISK
          | items EQ
          | items SET
          | items TABLE
          | items PARAM
          | NUM_VAL
          | WORD
          | STRING
          | QUOTEDSTRING
          | COMMA
          | COLON
          | LBRACKET
          | RBRACKET
          | LBRACE
          | RBRACE
          | TR
          | LPAREN
          | RPAREN
          | ASTERISK
          | EQ
          | SET
          | TABLE
          | PARAM
    '''
    # Locate and handle item as necessary
    single_item = len(p) == 2
    if single_item:
        tmp = p[1]
    else:
        tmp = p[2]
    if (
        type(tmp) is str
        and tmp[0] == '"'
        and tmp[-1] == '"'
        and len(tmp) > 2
        and not ' ' in tmp
    ):
        tmp = tmp[1:-1]

    # Grow items list according to parsed item length
    if single_item:
        p[0] = [tmp]
    else:
        # yacc __getitem__ is expensive: use a local list to avoid a
        # getitem call on p[0]
        tmp_lst = p[1]
        tmp_lst.append(tmp)
        p[0] = tmp_lst


def p_error(p):
    if p is None:
        tmp = "Syntax error at end of file."
    else:
        tmp = "Syntax error at token '%s' with value '%s' (line %s, column %s)" % (
            p.type,
            p.value,
            p.lineno,
            _lex_token_position(p),
        )
    raise IOError(tmp)


# --------------------------------------------------------------
# the DAT file lexer and yaccer only need to be
# created once, so have the corresponding objects
# accessible at module scope.
# --------------------------------------------------------------


tabmodule = '_parse_table_datacmds'

dat_lexer = None
dat_yaccer = None


def _get_this_file_signature():
    """Compute the SHA-256 signature for this module.

    We store the signature in the auto-generated parse table module so
    that we can robustly detect when the parse tables must be
    regenerated.

    """
    import hashlib

    sha = hashlib.sha256()
    with open(this_file(), 'r') as FILE:
        sha.update(FILE.read().encode())
    return sha.hexdigest()


#
# The function that performs the parsing
#
def parse_data_commands(data=None, filename=None, debug=0):
    global dat_lexer
    global dat_yaccer

    # if the lexer/yaccer haven't been initialized, do so.
    if dat_lexer is None or dat_yaccer is None:
        # Import the parse table, and check that it is valid for this module
        pkg = __name__.rsplit('.', maxsplit=1)[0]
        dat_tabmodule = importlib.import_module('.' + tabmodule, pkg)
        if _get_this_file_signature() != dat_tabmodule._lr_module_signature:
            raise DeveloperError(
                f"DAT parse tables ({pkg}.{tabmodule}) out of sync with "
                f"parser definition; regenerate by running {__file__}"
            )

        dat_lexer = lex.lex()
        dat_yaccer = yacc.yacc(tabmodule=dat_tabmodule, debug=debug)

    #
    # Load the text to parse
    #
    if filename is not None:
        if data is not None:
            raise ValueError(
                "parse_data_commands: cannot specify both "
                "data and filename arguments"
            )
        with open(filename, 'r') as FILE:
            data = FILE.read()

    if data is None:
        return None

    #
    # Initialize the lexer / parser objects
    #
    global _parse_info
    _parse_info = {None: []}
    dat_lexer.linepos = []
    #
    # Parse
    #
    dat_yaccer.parse(data, lexer=dat_lexer, debug=debug)
    #
    # Reset parse data and return the result
    #
    result = _parse_info
    _parse_info = None
    dat_lexer.linepos = []
    return result


if __name__ == '__main__':  # pragma:nocover
    print(f"Regenerating {tabmodule}")
    yacc.yacc(
        tabmodule=tabmodule,
        outputdir=this_file_dir(),
        debug=100,
        module_signature=_get_this_file_signature(),
    )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""
Test methods for construction and solution of master problem
objects.
"""

import logging
import time
import pyomo.common.unittest as unittest

from pyomo.common.collections import Bunch
from pyomo.common.dependencies import numpy_available, scipy_available
from pyomo.core.base import ConcreteModel, Constraint, minimize, Objective, Param, Var
from pyomo.core.expr import exp
from pyomo.core.expr.compare import assertExpressionsEqual
from pyomo.environ import SolverFactory
from pyomo.opt import TerminationCondition

from pyomo.contrib.pyros.uncertainty_sets import BoxSet
from pyomo.contrib.pyros.master_problem_methods import (
    add_scenario_block_to_master_problem,
    construct_initial_master_problem,
    construct_master_feasibility_problem,
    construct_dr_polishing_problem,
    MasterProblemData,
    higher_order_decision_rule_efficiency,
)
from pyomo.contrib.pyros.util import (
    ModelData,
    preprocess_model_data,
    get_all_first_stage_eq_cons,
    ObjectiveType,
    time_code,
    TimingData,
    VariablePartitioning,
    pyrosTerminationCondition,
)

if not (numpy_available and scipy_available):
    raise unittest.SkipTest("Packages numpy and scipy must both be available.")

_baron = SolverFactory("baron")
baron_available = _baron.available()
baron_license_is_valid = _baron.license_is_valid()


logger = logging.getLogger(__name__)


def build_simple_model_data(objective_focus="worst_case", decision_rule_order=1):
    """
    Test construction of master problem.
    """
    m = ConcreteModel()
    m.u = Param(initialize=0.5, mutable=True)
    m.x1 = Var(bounds=[-1000, 1000], initialize=1)
    m.x2 = Var(bounds=[-1000, 1000], initialize=1)
    m.x3 = Var(bounds=[-1000, 1000], initialize=-3)
    m.con = Constraint(expr=exp(m.u - 1) - m.x1 - m.x2 * m.u - m.x3 * m.u**2 <= 0)
    m.eq_con = Constraint(expr=m.x2 - 1 == 0)

    m.obj = Objective(expr=m.x1 + m.x2 / 2 + m.x3 / 3)

    config = Bunch(
        uncertain_params=[m.u],
        objective_focus=ObjectiveType[objective_focus],
        decision_rule_order=decision_rule_order,
        progress_logger=logger,
        nominal_uncertain_param_vals=[0.4],
        separation_priority_order=dict(),
        uncertainty_set=BoxSet([[1, 2]]),
    )
    model_data = ModelData(original_model=m, timing=TimingData(), config=config)
    user_var_partitioning = VariablePartitioning(
        first_stage_variables=[m.x1],
        second_stage_variables=[m.x2, m.x3],
        state_variables=[],
    )

    preprocess_model_data(model_data, user_var_partitioning)

    return model_data


class TestConstructMasterProblem(unittest.TestCase):
    """
    Tests for construction of the master problem and
    scenario sub-blocks.
    """

    def test_initial_construct_master(self):
        """
        Test initial construction of the master problem
        from the preprocesed working model.
        """
        model_data = build_simple_model_data()
        master_model = construct_initial_master_problem(model_data)

        self.assertTrue(hasattr(master_model, "scenarios"))
        self.assertIsNot(master_model.scenarios[0, 0], model_data.working_model)
        self.assertTrue(master_model.epigraph_obj.active)
        self.assertIs(
            master_model.epigraph_obj.expr,
            master_model.scenarios[0, 0].first_stage.epigraph_var,
        )

        # check all the variables (including first-stage ones)
        # were cloned
        nadj_var_zip = zip(
            master_model.scenarios[0, 0].all_nonadjustable_variables,
            model_data.working_model.all_nonadjustable_variables,
        )
        for master_var, wm_var in nadj_var_zip:
            self.assertIsNot(
                master_var,
                wm_var,
                f"Variable with name {wm_var.name!r} not cloned as expected.",
            )

        # check parameter value is set to the nominal realization
        self.assertEqual(
            master_model.scenarios[0, 0].user_model.u.value,
            model_data.config.nominal_uncertain_param_vals[0],
        )

    def test_add_scenario_block_to_master(self):
        """
        Test method for adding scenario block to an already
        constructed master problem, without cloning of the
        first-stage variables.
        """
        model_data = build_simple_model_data()
        master_model = construct_initial_master_problem(model_data)
        add_scenario_block_to_master_problem(
            master_model=master_model,
            scenario_idx=[0, 1],
            param_realization=[0.6],
            from_block=master_model.scenarios[0, 0],
            clone_first_stage_components=False,
        )

        self.assertEqual(master_model.scenarios[0, 1].user_model.u.value, 0.6)

        nadj_var_zip = zip(
            master_model.scenarios[0, 0].all_nonadjustable_variables,
            master_model.scenarios[0, 1].all_nonadjustable_variables,
        )
        for var_00, var_01 in nadj_var_zip:
            self.assertIs(
                var_00,
                var_01,
                msg=f"Variable {var_00.name} was cloned across scenario blocks.",
            )

        # the first-stage inequality and equality constraints
        # should be cloned. we do this to avoid issues with the solver
        # interfaces (such as issues with manipulating symbol maps)
        nadj_ineq_con_zip = zip(
            master_model.scenarios[0, 0].first_stage.inequality_cons.values(),
            master_model.scenarios[0, 1].first_stage.inequality_cons.values(),
        )
        for ineq_con_00, ineq_con_01 in nadj_ineq_con_zip:
            self.assertIsNot(
                ineq_con_00,
                ineq_con_01,
                msg=(
                    f"first-stage inequality con {ineq_con_00.name!r} was not "
                    "cloned across scenario blocks."
                ),
            )
            self.assertTrue(
                ineq_con_00.active,
                msg=(
                    "First-stage inequality constraint "
                    f"{ineq_con_00.name!r} should be active."
                ),
            )
            self.assertFalse(
                ineq_con_01.active,
                msg=(
                    "Duplicate first-stage inequality constraint "
                    f"{ineq_con_01.name!r} should be deactivated"
                ),
            )

        nadj_eq_con_zip = zip(
            get_all_first_stage_eq_cons(master_model.scenarios[0, 0]),
            get_all_first_stage_eq_cons(master_model.scenarios[0, 1]),
        )
        for eq_con_00, eq_con_01 in nadj_eq_con_zip:
            self.assertIsNot(
                eq_con_00,
                eq_con_01,
                msg=(
                    f"first-stage equality con {eq_con_00.name} was not cloned "
                    "across scenario blocks."
                ),
            )
            self.assertTrue(
                eq_con_00.active,
                msg=(
                    "First-stage equality constraint "
                    f"{eq_con_00.name!r} should be active."
                ),
            )
            self.assertFalse(
                eq_con_01.active,
                msg=(
                    "Duplicate first-stage equality constraint "
                    f"{eq_con_01.name!r} should be deactivated"
                ),
            )


class TestNewConstructMasterFeasibilityProblem(unittest.TestCase):
    """
    Test construction of the master feasibility problem.
    """

    def build_simple_master_data(self):
        """
        Construct master data-like object for feasibility problem
        tests.
        """
        model_data = build_simple_model_data()
        master_model = construct_initial_master_problem(model_data)
        add_scenario_block_to_master_problem(
            master_model=master_model,
            scenario_idx=[1, 0],
            param_realization=[1],
            from_block=master_model.scenarios[0, 0],
            clone_first_stage_components=False,
        )
        master_data = Bunch(
            master_model=master_model, iteration=1, config=model_data.config
        )

        return master_data

    def test_construct_master_feasibility_problem_var_map(self):
        """
        Test construction of feasibility problem var map.
        """
        master_data = self.build_simple_master_data()
        slack_model = construct_master_feasibility_problem(master_data)

        self.assertTrue(master_data.feasibility_problem_varmap)
        for mvar, feasvar in master_data.feasibility_problem_varmap:
            self.assertIs(
                mvar,
                master_data.master_model.find_component(feasvar),
                msg=f"{mvar.name!r} is not same as find_component({feasvar.name!r})",
            )
            self.assertIs(
                feasvar,
                slack_model.find_component(mvar),
                msg=f"{feasvar.name!r} is not same as find_component({mvar.name!r})",
            )

    def test_construct_master_feasibility_problem_slack_vars(self):
        """
        Check master feasibility slack variables.
        """
        master_data = self.build_simple_master_data()
        slack_model = construct_master_feasibility_problem(master_data)

        slack_var_blk = slack_model._core_add_slack_variables
        scenario_10_blk = slack_model.scenarios[1, 0]

        # test a few of the constraints
        slack_user_model_x3_lb_con = scenario_10_blk.second_stage.inequality_cons[
            "var_x3_certain_lower_bound_con"
        ]
        slack_user_model_x3_lb_con_var = slack_var_blk.find_component(
            "'_slack_minus_scenarios[1,0].second_stage.inequality_cons["
            "var_x3_certain_lower_bound_con]'"
        )
        assertExpressionsEqual(
            self,
            slack_user_model_x3_lb_con.body <= slack_user_model_x3_lb_con.upper,
            -scenario_10_blk.user_model.x3 - slack_user_model_x3_lb_con_var <= 1000.0,
        )
        self.assertEqual(slack_user_model_x3_lb_con_var.value, 0)

        slack_user_model_x3_ub_con = scenario_10_blk.second_stage.inequality_cons[
            "var_x3_certain_upper_bound_con"
        ]
        slack_user_model_x3_ub_con_var = slack_var_blk.find_component(
            "'_slack_minus_scenarios[1,0].second_stage.inequality_cons["
            "var_x3_certain_upper_bound_con]'"
        )
        assertExpressionsEqual(
            self,
            slack_user_model_x3_ub_con.body <= slack_user_model_x3_ub_con.upper,
            scenario_10_blk.user_model.x3 - slack_user_model_x3_ub_con_var <= 1000.0,
        )
        self.assertEqual(slack_user_model_x3_lb_con_var.value, 0)

        # constraint 'con' is violated when u = 0.8;
        # check slack initialization
        slack_user_model_con_var = slack_var_blk.find_component(
            "'_slack_minus_scenarios[1,0].second_stage.inequality_cons"
            "[ineq_con_con_upper_bound_con]'"
        )
        self.assertEqual(
            slack_user_model_con_var.value,
            -master_data.master_model.scenarios[1, 0].user_model.con.uslack(),
        )

    def test_construct_master_feasibility_problem_obj(self):
        """
        Check master feasibility slack variables.
        """
        master_data = self.build_simple_master_data()
        slack_model = construct_master_feasibility_problem(master_data)

        self.assertFalse(slack_model.epigraph_obj.active)
        self.assertTrue(slack_model._core_add_slack_variables._slack_objective.active)


class TestDRPolishingProblem(unittest.TestCase):
    """
    Tests for the PyROS DR polishing problem.
    """

    def build_simple_master_data(self):
        """
        Construct master data-like object for feasibility problem
        tests.
        """
        model_data = build_simple_model_data()
        master_model = construct_initial_master_problem(model_data)
        add_scenario_block_to_master_problem(
            master_model=master_model,
            scenario_idx=[1, 0],
            param_realization=[0.1],
            from_block=master_model.scenarios[0, 0],
            clone_first_stage_components=False,
        )
        master_data = Bunch(
            master_model=master_model, iteration=1, config=model_data.config
        )

        return master_data

    def test_construct_dr_polishing_problem_nonadj_components(self):
        """
        Test state of the nonadjustable components
        of the DR polishing problem.
        """
        master_data = self.build_simple_master_data()
        polishing_model = construct_dr_polishing_problem(master_data)
        eff_first_stage_vars = polishing_model.scenarios[
            0, 0
        ].effective_var_partitioning.first_stage_variables
        for effective_first_stage_var in eff_first_stage_vars:
            self.assertTrue(
                effective_first_stage_var.fixed,
                msg=(
                    "Effective first-stage variable "
                    f"{effective_first_stage_var.name!r} "
                    "not fixed."
                ),
            )

        nom_polishing_block = polishing_model.scenarios[0, 0]
        self.assertTrue(nom_polishing_block.first_stage.epigraph_var.fixed)
        self.assertFalse(nom_polishing_block.first_stage.decision_rule_vars[0][0].fixed)
        self.assertFalse(nom_polishing_block.first_stage.decision_rule_vars[0][1].fixed)

        # ensure constraints in fixed vars were deactivated
        self.assertFalse(nom_polishing_block.user_model.eq_con.active)

        # these have either unfixed DR or adjustable variables,
        # so they should remain active
        # self.assertTrue(nom_polishing_block.user_model.con.active)
        self.assertTrue(
            nom_polishing_block.second_stage.inequality_cons[
                "ineq_con_con_upper_bound_con"
            ].active
        )
        self.assertTrue(nom_polishing_block.second_stage.decision_rule_eqns[0].active)

    def test_construct_dr_polishing_problem_polishing_components(self):
        """
        Test auxiliary Var/Constraint components of the DR polishing
        problem.
        """
        master_data = self.build_simple_master_data()
        # DR order is 1, and x3 is second-stage.
        # to test fixing efficiency, fix the affine DR variable
        decision_rule_vars = master_data.master_model.scenarios[
            0, 0
        ].first_stage.decision_rule_vars
        decision_rule_vars[0][1].fix()
        polishing_model = construct_dr_polishing_problem(master_data)
        nom_polishing_block = polishing_model.scenarios[0, 0]

        self.assertFalse(decision_rule_vars[0][0].fixed)
        self.assertTrue(polishing_model.polishing_vars[0][0].fixed)
        self.assertFalse(polishing_model.polishing_abs_val_lb_con_0[0].active)
        self.assertFalse(polishing_model.polishing_abs_val_ub_con_0[0].active)

        # polishing components for the affine DR term should be
        # fixed/deactivated since the DR variable was fixed
        self.assertTrue(decision_rule_vars[0][1].fixed)
        self.assertTrue(polishing_model.polishing_vars[0][1].fixed)
        self.assertFalse(polishing_model.polishing_abs_val_lb_con_0[1].active)
        self.assertFalse(polishing_model.polishing_abs_val_ub_con_0[1].active)

        # check initialization of polishing vars
        self.assertEqual(
            polishing_model.polishing_vars[0][0].value,
            abs(nom_polishing_block.first_stage.decision_rule_vars[0][0].value),
        )
        self.assertEqual(
            polishing_model.polishing_vars[0][1].value,
            abs(nom_polishing_block.first_stage.decision_rule_vars[0][1].value),
        )

        assertExpressionsEqual(
            self,
            polishing_model.polishing_obj.expr,
            polishing_model.polishing_vars[0][0] + polishing_model.polishing_vars[0][1],
        )
        self.assertEqual(polishing_model.polishing_obj.sense, minimize)

    def test_construct_dr_polishing_problem_objectives(self):
        """
        Test states of the Objective components of the DR
        polishing model.
        """
        master_data = self.build_simple_master_data()
        polishing_model = construct_dr_polishing_problem(master_data)
        self.assertFalse(polishing_model.epigraph_obj.active)
        self.assertTrue(polishing_model.polishing_obj.active)

    def test_construct_dr_polishing_problem_params_zero(self):
        """
        Check that DR polishing fixes/deactivates components
        for DR expression terms where the product of uncertain
        parameters is below tolerance.
        """
        master_data = self.build_simple_master_data()

        # trigger fixing of the corresponding polishing vars
        master_data.master_model.scenarios[0, 0].user_model.u.set_value(1e-10)
        master_data.master_model.scenarios[1, 0].user_model.u.set_value(1e-11)

        polishing_model = construct_dr_polishing_problem(master_data)

        dr_vars = polishing_model.scenarios[0, 0].first_stage.decision_rule_vars

        # since static DR terms should not be polished
        self.assertTrue(polishing_model.polishing_vars[0][0].fixed)
        self.assertFalse(polishing_model.polishing_abs_val_lb_con_0[0].active)
        self.assertFalse(polishing_model.polishing_abs_val_ub_con_0[0].active)

        # affine term should be fixed to 0,
        # since the uncertain param values are small enough.
        # polishing constraints are deactivated since we don't need them
        self.assertTrue(dr_vars[0][1].fixed)
        self.assertEqual(dr_vars[0][1].value, 0)
        self.assertTrue(polishing_model.polishing_vars[0][1].fixed)
        self.assertFalse(polishing_model.polishing_abs_val_lb_con_0[1].active)
        self.assertFalse(polishing_model.polishing_abs_val_ub_con_0[1].active)


class TestHigherOrderDecisionRuleEfficiency(unittest.TestCase):
    """
    Test efficiency for decision rules.
    """

    def test_higher_order_decision_rule_efficiency(self):
        """
        Test higher-order decision rule efficiency.
        """
        model_data = build_simple_model_data(decision_rule_order=2)
        master_model = construct_initial_master_problem(model_data)
        master_data = Bunch(
            master_model=master_model, iteration=0, config=model_data.config
        )
        decision_rule_vars = master_data.master_model.scenarios[
            0, 0
        ].first_stage.decision_rule_vars[0]

        for iter_num in range(4):
            master_data.iteration = iter_num
            higher_order_decision_rule_efficiency(master_data)
            self.assertFalse(
                decision_rule_vars[0].fixed,
                msg=(
                    f"DR Var {decision_rule_vars[1].name!r} should not "
                    f"be fixed by efficiency in iteration {iter_num}"
                ),
            )
            if iter_num == 0:
                self.assertTrue(
                    decision_rule_vars[1].fixed,
                    msg=(
                        f"DR Var {decision_rule_vars[1].name!r} should "
                        f"be fixed by efficiency in iteration {iter_num}"
                    ),
                )
                self.assertTrue(
                    decision_rule_vars[2].fixed,
                    msg=(
                        f"DR Var {decision_rule_vars[2].name!r} should "
                        f"be fixed by efficiency in iteration {iter_num}"
                    ),
                )
            elif iter_num <= len(master_data.config.uncertain_params):
                self.assertFalse(
                    decision_rule_vars[1].fixed,
                    msg=(
                        f"DR Var {decision_rule_vars[1].name!r} should not "
                        f"be fixed by efficiency in iteration {iter_num}"
                    ),
                )
                self.assertTrue(
                    decision_rule_vars[2].fixed,
                    msg=(
                        f"DR Var {decision_rule_vars[2].name!r} should "
                        f"be fixed by efficiency in iteration {iter_num}"
                    ),
                )
            else:
                self.assertFalse(
                    decision_rule_vars[1].fixed,
                    msg=(
                        f"DR Var {decision_rule_vars[1].name!r} should not "
                        f"be fixed by efficiency in iteration {iter_num}"
                    ),
                )
                self.assertFalse(
                    decision_rule_vars[2].fixed,
                    msg=(
                        f"DR Var {decision_rule_vars[2].name!r} should not "
                        f"be fixed by efficiency in iteration {iter_num}"
                    ),
                )


class TestSolveMaster(unittest.TestCase):
    """
    Test method for solving master problem
    """

    @unittest.skipUnless(baron_available, "Global NLP solver is not available.")
    def test_solve_master(self):
        model_data = build_simple_model_data()
        model_data.timing = TimingData()
        baron = SolverFactory("baron")
        model_data.config.update(
            dict(
                local_solver=baron,
                global_solver=baron,
                backup_local_solvers=[],
                backup_global_solvers=[],
                tee=False,
            )
        )
        master_data = MasterProblemData(model_data)
        with time_code(master_data.timing, "main", is_main_timer=True):
            master_soln = master_data.solve_master()
            self.assertEqual(len(master_soln.master_results_list), 1)
            self.assertIsNone(master_soln.feasibility_problem_results)
            self.assertIsNone(master_soln.pyros_termination_condition)
            self.assertIs(master_soln.master_model, master_data.master_model)
            self.assertEqual(
                master_soln.master_results_list[0].solver.termination_condition,
                TerminationCondition.optimal,
                msg=(
                    "Could not solve simple master problem with solve_master "
                    "function."
                ),
            )

    @unittest.skipUnless(baron_available, "Global NLP solver is not available")
    def test_solve_master_timeout_on_master(self):
        """
        Test method for solution of master problems times out
        on feasibility problem.
        """
        model_data = build_simple_model_data()
        model_data.timing = TimingData()
        baron = SolverFactory("baron")
        model_data.config.update(
            dict(
                local_solver=baron,
                global_solver=baron,
                backup_local_solvers=[],
                backup_global_solvers=[],
                tee=False,
                time_limit=1,
            )
        )
        master_data = MasterProblemData(model_data)
        with time_code(master_data.timing, "main", is_main_timer=True):
            time.sleep(1)
            master_soln = master_data.solve_master()
            self.assertIsNone(master_soln.feasibility_problem_results)
            self.assertEqual(master_soln.master_model, master_data.master_model)
            self.assertEqual(len(master_soln.master_results_list), 1)
            self.assertEqual(
                master_soln.master_results_list[0].solver.termination_condition,
                TerminationCondition.optimal,
                msg=(
                    "Could not solve simple master problem with solve_master "
                    "function."
                ),
            )
            self.assertEqual(
                master_soln.pyros_termination_condition,
                pyrosTerminationCondition.time_out,
            )

    @unittest.skipUnless(baron_available, "Global NLP solver is not available")
    def test_solve_master_timeout_on_master_feasibility(self):
        """
        Test method for solution of master problems times out
        on feasibility problem.
        """
        model_data = build_simple_model_data()
        model_data.timing = TimingData()
        baron = SolverFactory("baron")
        model_data.config.update(
            dict(
                local_solver=baron,
                global_solver=baron,
                backup_local_solvers=[],
                backup_global_solvers=[],
                tee=False,
                time_limit=1,
            )
        )
        master_data = MasterProblemData(model_data)
        add_scenario_block_to_master_problem(
            master_data.master_model,
            scenario_idx=[1, 0],
            param_realization=[0.6],
            from_block=master_data.master_model.scenarios[0, 0],
            clone_first_stage_components=False,
        )
        master_data.iteration = 1
        with time_code(master_data.timing, "main", is_main_timer=True):
            time.sleep(1)
            master_soln = master_data.solve_master()
            self.assertIsNotNone(master_soln.feasibility_problem_results)
            self.assertFalse(master_soln.master_results_list)
            self.assertIs(master_soln.master_model, master_data.master_model)
            self.assertEqual(
                master_soln.pyros_termination_condition,
                pyrosTerminationCondition.time_out,
            )


class TestPolishDRVars(unittest.TestCase):
    """
    Test DR polishing subroutine.
    """

    @unittest.skipUnless(
        baron_license_is_valid, "Global NLP solver is not available and licensed."
    )
    def test_polish_dr_vars(self):
        model_data = build_simple_model_data()
        model_data.timing = TimingData()
        baron = SolverFactory("baron")
        model_data.config.update(
            dict(
                local_solver=baron,
                global_solver=baron,
                backup_local_solvers=[],
                backup_global_solvers=[],
                tee=False,
            )
        )
        master_data = MasterProblemData(model_data)
        add_scenario_block_to_master_problem(
            master_data.master_model,
            scenario_idx=[1, 0],
            param_realization=[0.6],
            from_block=master_data.master_model.scenarios[0, 0],
            clone_first_stage_components=False,
        )
        master_data.iteration = 1

        master_data.timing = TimingData()
        with time_code(master_data.timing, "main", is_main_timer=True):
            master_soln = master_data.solve_master()
            self.assertEqual(
                master_soln.master_results_list[0].solver.termination_condition,
                TerminationCondition.optimal,
            )

            results, success = master_data.solve_dr_polishing()
            self.assertEqual(
                results.solver.termination_condition,
                TerminationCondition.optimal,
                msg="Minimize dr norm did not solve to optimality.",
            )
            self.assertTrue(
                success, msg=f"DR polishing success {success}, expected True."
            )


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
import pytest

from pyomo.contrib.mpc.data.find_nearest_index import (
    find_nearest_index,
    find_nearest_interval_index,
)


class TestFindNearestIndex(unittest.TestCase):
    def test_two_points(self):
        array = [0, 5]

        i = find_nearest_index(array, 1)
        self.assertEqual(i, 0)
        i = find_nearest_index(array, 1, tolerance=0.5)
        self.assertEqual(i, None)

        i = find_nearest_index(array, -0.01, tolerance=0.1)
        self.assertEqual(i, 0)
        i = find_nearest_index(array, -0.01, tolerance=0.001)
        self.assertEqual(i, None)

        i = find_nearest_index(array, 6, tolerance=2)
        self.assertEqual(i, 1)
        i = find_nearest_index(array, 6, tolerance=1)
        self.assertEqual(i, 1)

        # This test relies on the behavior for tiebreaks
        i = find_nearest_index(array, 2.5)
        self.assertEqual(i, 0)

    def test_array_with_floats(self):
        array = []
        for i in range(5):
            i0 = float(i)
            i1 = round((i + 0.15) * 1e4) / 1e4
            i2 = round((i + 0.64) * 1e4) / 1e4
            array.extend([i, i1, i2])
        array.append(5.0)

        i = find_nearest_index(array, 1.01, tolerance=0.1)
        self.assertEqual(i, 3)
        i = find_nearest_index(array, 1.01, tolerance=0.001)
        self.assertEqual(i, None)

        i = find_nearest_index(array, 3.5)
        self.assertEqual(i, 11)
        i = find_nearest_index(array, 3.5, tolerance=0.1)
        self.assertEqual(i, None)

        i = find_nearest_index(array, -1)
        self.assertEqual(i, 0)
        i = find_nearest_index(array, -1, tolerance=1)
        self.assertEqual(i, 0)

        i = find_nearest_index(array, 5.5)
        self.assertEqual(i, 15)
        i = find_nearest_index(array, 5.5, tolerance=0.49)
        self.assertEqual(i, None)

        i = find_nearest_index(array, 2.64, tolerance=1e-8)
        self.assertEqual(i, 8)
        i = find_nearest_index(array, 2.64, tolerance=0)
        self.assertEqual(i, 8)

        i = find_nearest_index(array, 5, tolerance=0)
        self.assertEqual(i, 15)

        i = find_nearest_index(array, 0, tolerance=0)
        self.assertEqual(i, 0)


class TestFindNearestIntervalIndex(unittest.TestCase):
    def test_find_interval(self):
        intervals = [(0.0, 0.1), (0.1, 0.5), (0.7, 1.0)]
        target = 0.05
        idx = find_nearest_interval_index(intervals, target)
        self.assertEqual(idx, 0)

        target = 0.099
        idx = find_nearest_interval_index(intervals, target)
        self.assertEqual(idx, 0)

        target = 0.1
        idx = find_nearest_interval_index(intervals, target)
        self.assertEqual(idx, 0)

        target = 0.1
        idx = find_nearest_interval_index(intervals, target, prefer_left=False)
        self.assertEqual(idx, 1)

        target = 0.55
        idx = find_nearest_interval_index(intervals, target)
        self.assertEqual(idx, 1)

        target = 0.60
        idx = find_nearest_interval_index(intervals, target)
        self.assertEqual(idx, 1)

        target = 0.6999
        idx = find_nearest_interval_index(intervals, target)
        self.assertEqual(idx, 2)

        target = 1.0
        idx = find_nearest_interval_index(intervals, target)
        self.assertEqual(idx, 2)

        target = -0.1
        idx = find_nearest_interval_index(intervals, target)
        self.assertEqual(idx, 0)

        target = 1.1
        idx = find_nearest_interval_index(intervals, target)
        self.assertEqual(idx, 2)

    def test_find_interval_tolerance(self):
        intervals = [(0.0, 0.1), (0.1, 0.5), (0.7, 1.0)]

        target = 0.501
        idx = find_nearest_interval_index(intervals, target, tolerance=None)
        self.assertEqual(idx, 1)

        idx = find_nearest_interval_index(intervals, target, tolerance=1e-5)
        self.assertEqual(idx, None)

        idx = find_nearest_interval_index(intervals, target, tolerance=1e-2)
        self.assertEqual(idx, 1)

        target = 1.001
        idx = find_nearest_interval_index(intervals, target, tolerance=1e-2)
        self.assertEqual(idx, 2)

        #
        # Behavior when distance between target and nearest interval "equals"
        # the tolerance is not well-defined. Here the computed distance may
        # not be exactly 1e-3 due to roundoff error.
        #
        # idx = find_nearest_interval_index(intervals, target, tolerance=1e-3)
        # self.assertEqual(idx, None)

        idx = find_nearest_interval_index(intervals, target, tolerance=1e-4)
        self.assertEqual(idx, None)

    def test_find_interval_with_tolerance_on_boundary(self):
        # Our target is on the boundary between two intervals.
        intervals = [(0.0, 0.1), (0.1, 0.5), (0.5, 1.0)]
        target = 0.1001
        idx = find_nearest_interval_index(
            intervals, target, tolerance=None, prefer_left=True
        )
        self.assertEqual(idx, 1)

        # target != 0.1 (the interval boundary) within tolerance. We are
        # within interval 1.
        idx = find_nearest_interval_index(
            intervals, target, tolerance=1e-5, prefer_left=True
        )
        self.assertEqual(idx, 1)
        # This is true even if we prefer the right interval
        idx = find_nearest_interval_index(
            intervals, target, tolerance=1e-5, prefer_left=False
        )
        self.assertEqual(idx, 1)

        # target == 0.1 within tolerance. We are on the boundary, and
        # should return the "preferred" interval.
        idx = find_nearest_interval_index(
            intervals, target, tolerance=1e-3, prefer_left=True
        )
        self.assertEqual(idx, 0)
        idx = find_nearest_interval_index(
            intervals, target, tolerance=1e-3, prefer_left=False
        )
        self.assertEqual(idx, 1)

        target = 0.4999
        # We are not equal to boundary (0.5) within tolerance
        idx = find_nearest_interval_index(
            intervals, target, tolerance=1e-5, prefer_left=True
        )
        self.assertEqual(idx, 1)
        idx = find_nearest_interval_index(
            intervals, target, tolerance=1e-5, prefer_left=False
        )
        self.assertEqual(idx, 1)

        # We are equal to boundary within tolerance
        idx = find_nearest_interval_index(
            intervals, target, tolerance=1e-3, prefer_left=True
        )
        self.assertEqual(idx, 1)
        idx = find_nearest_interval_index(
            intervals, target, tolerance=1e-3, prefer_left=False
        )
        self.assertEqual(idx, 2)

    def test_find_interval_with_tolerance_singleton(self):
        intervals = [(0.0, 0.1), (0.1, 0.1), (0.5, 0.5), (0.5, 1.0)]

        target = 0.1001
        idx = find_nearest_interval_index(
            intervals, target, tolerance=1e-3, prefer_left=True
        )
        self.assertEqual(idx, 0)
        idx = find_nearest_interval_index(
            intervals, target, tolerance=1e-3, prefer_left=False
        )
        self.assertEqual(idx, 1)

        target = 0.0999
        idx = find_nearest_interval_index(
            intervals, target, tolerance=1e-3, prefer_left=True
        )
        self.assertEqual(idx, 0)
        idx = find_nearest_interval_index(
            intervals, target, tolerance=1e-3, prefer_left=False
        )
        self.assertEqual(idx, 1)

        target = 0.4999
        idx = find_nearest_interval_index(
            intervals, target, tolerance=1e-3, prefer_left=True
        )
        self.assertEqual(idx, 2)
        idx = find_nearest_interval_index(
            intervals, target, tolerance=1e-3, prefer_left=False
        )
        self.assertEqual(idx, 3)

        target = 0.5001
        idx = find_nearest_interval_index(
            intervals, target, tolerance=1e-3, prefer_left=True
        )
        self.assertEqual(idx, 2)
        idx = find_nearest_interval_index(
            intervals, target, tolerance=1e-3, prefer_left=False
        )
        self.assertEqual(idx, 3)


if __name__ == "__main__":
    unittest.main()
